# -*- coding: utf-8 -*-
"""
Created on Sat Feb 15 18:15:16 2025

@author: dvezinet
"""


import copy


import numpy as np
import scipy.integrate as scpinteg
import datastock as ds


# #############################################################
# #############################################################
#                  DEFAULT
# #############################################################


_DSOLVER = {
    'method': 'RK45',
    'first_step': None,
    'min_step': 0.,
    'max_step': np.inf,
    'rtol': 1e-6,
    'atol': 1e-6,
    'dense_output': False,
    'vectorized': False,   # only for ‘Radau’ and ‘BDF’
    'fac': None,           # only for ‘Radau’, ‘BDF’ and ‘LSODA’
}


# #############################################################
# #############################################################
#                  Main
# #############################################################


def main(
    coll=None,
    # 3 componants
    key_XR=None,
    key_YZ=None,
    key_Zphi=None,
    # domain
    domain=None,
    # linear vs toroidal
    geometry=None,
    # starting points
    pts_X=None,
    pts_Y=None,
    pts_Z=None,
    # length and resolution of lines
    res=None,
    length=None,
    npts=None,
    dl=None,
    # direction
    direct=None,
    # solver options
    dsolver=None,
):

    # ------------------
    # check inputs
    # ------------------

    (
        dkey,
        dpts,
        geometry,
        dl,
        direct,
        dsolver,
    ) = _check(
        coll=coll,
        # 3 componants
        key_XR=key_XR,
        key_YZ=key_YZ,
        key_Zphi=key_Zphi,
        # linear vs toroidal
        geometry=geometry,
        # starting points
        pts_X=pts_X,
        pts_Y=pts_Y,
        pts_Z=pts_Z,
        # length and resolution of lines
        res=res,
        length=length,
        npts=npts,
        dl=dl,
        # direction
        direct=direct,
        # solver
        dsolver=dsolver,
    )

    # ------------------
    # prepare output
    # ------------------

    dout, update_sli_dom, shape_key_nobs = _prepare_output(
        coll=coll,
        dkey=dkey,
        dpts=dpts,
        dl=dl,
        direct=direct,
    )

    # ------------------
    # get derivative functions
    # ------------------

    func_fwd, func_bck = _get_func(
        coll=coll,
        dkey=dkey,
        geometry=geometry,
    )

    # ------------------
    # compute
    # ------------------

    # -----------------------
    # loop on starting points

    for i0, ipts in enumerate(zip(*dpts['iok'].nonzero())):

        # ------------------
        # loop on field refs

        for i1, iref in enumerate(np.ndindex(shape_key_nobs)):

            # update domain
            (
                sli_fwd, sli_bck,
                domaini,
                sli_ev,
            ) = update_sli_dom(ipts, iref)

            # -------
            # Forward

            if 'fwd' in direct:

                sol = scpinteg.solve_ivp(
                    func_fwd,
                    [dl[0], dl[-1]],
                    [pts_X[ipts], pts_Y[ipts], pts_Z[ipts]],
                    args=(domaini,),
                    t_eval=dl,
                    **dsolver,
                )

                dout['nfev_fwd']['data'][sli_ev] = sol.nfev
                dout['status_fwd']['data'][sli_ev] = sol.status

                if sol.success is True:
                    dout['pts_X']['data'][sli_fwd] = sol.y[0, :]
                    dout['pts_Y']['data'][sli_fwd] = sol.y[1, :]
                    dout['pts_Z']['data'][sli_fwd] = sol.y[2, :]

            # --------
            # Backward

            if 'bck' in direct:

                sol = scpinteg.solve_ivp(
                    func_bck,
                    [dl[0], dl[-1]],
                    [pts_X[ipts], pts_Y[ipts], pts_Z[ipts]],
                    args=(domaini,),
                    t_eval=dl,
                    **dsolver,
                )

                dout['nfev_bck']['data'][sli_ev] = sol.nfev
                dout['status_bck']['data'][sli_ev] = sol.status

                if sol.success is True:
                    dout['pts_X']['data'][sli_bck] = sol.y[0, :]
                    dout['pts_Y']['data'][sli_bck] = sol.y[1, :]
                    dout['pts_Z']['data'][sli_bck] = sol.y[2, :]

    # ------------------
    # return
    # ------------------

    if geometry == 'toroidal':
        _add_dphi(
            dout,
            direct=direct,
            npts=dl.size,
        )

    return dout


# #############################################################
# #############################################################
#                  Check
# #############################################################


def _check(
    coll=None,
    # 3 componants
    key_XR=None,
    key_YZ=None,
    key_Zphi=None,
    # linear vs toroidal
    geometry=None,
    # starting points
    pts_X=None,
    pts_Y=None,
    pts_Z=None,
    # length and resolution of lines
    res=None,
    length=None,
    npts=None,
    dl=None,
    # direction
    direct=None,
    # solver
    dsolver=None,
):

    # ------------------
    # key coordinates
    # ------------------

    dkey = _check_keys_components(
        coll=coll,
        # 3 componants
        key_XR=key_XR,
        key_YZ=key_YZ,
        key_Zphi=key_Zphi,
    )

    # ------------------
    # geometry
    # ------------------

    geometry = ds._generic_check._check_var(
        geometry, 'geometry',
        types=str,
        default='toroidal',
        allowed=['toroidal', 'linear'],
    )

    # ------------------
    # starting points
    # ------------------

    dpts = _check_pts(
        pts_X=pts_X,
        pts_Y=pts_Y,
        pts_Z=pts_Z,
    )

    # ------------------
    # resolution and length
    # ------------------

    dl = _check_res_length(
        res=res,
        length=length,
        npts=npts,
        dl=dl,
    )

    # ------------------
    # direction
    # ------------------

    if isinstance(direct, str):
        direct = [direct]

    lok = ['fwd', 'bck']
    direct = ds._generic_check._check_var_iter(
        direct, 'direct',
        types=(list, tuple),
        types_iter=str,
        default=lok,
        allowed=lok,
    )

    # ------------------
    # solver
    # ------------------

    dsolver = _check_dsolver(dsolver)

    return (
        dkey,
        dpts,
        geometry,
        dl,
        direct,
        dsolver,
    )


# #############################################################
# #############################################################
#                  Check components
# #############################################################


def _check_keys_components(
    coll=None,
    # 3 componants
    key_XR=None,
    key_YZ=None,
    key_Zphi=None,
):

    # --------------
    # prepare
    # --------------

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    # check valid 2d bsplines exist
    lok_bs = [
        k0 for k0, v0 in coll.dobj.get(wbs, {}).items()
        if coll.dobj[wm][v0[wm]]['nd'] == '2d'
    ]
    if len(lok_bs) == 0:
        msg = (
            "Line tracing for vector field: no valid bsplines2d!\n\n"
            + coll.show(wbs, verb=False, returnas=str)
        )
        raise Exception(msg)

    # check valid data defined on 2d bspline exist
    lok_data = [
        k0 for k0, v0 in coll.ddata.items()
        if v0.get(wbs) is not None
        and any([bs in lok_bs for bs in v0[wbs]])
    ]
    if len(lok_data) == 0:
        msg = (
            "Line tracing for vector field:"
            " no valid data defined on 2d bsplines\n"
            f"\t- lok_bs = {lok_bs}\n"
        )
        raise Exception(msg)

    # -------------
    # initialize
    # -------------

    dkey = {'key_XR': key_XR, 'key_YZ': key_YZ, 'key_Zphi': key_Zphi}
    dbs = {}
    dref = {}

    # -----------------
    # check component by component
    # -----------------

    for k0, v0 in dkey.items():

        c0 = v0 in lok_data and len(coll.ddata[v0][wbs]) == 1
        if not c0:
            msg = (
                "Line tracing for vector field wrong component '{k0}'!\n"
                f"\t- Not dependent on a (single) valid 2d bsplines!\n"
                f"\t- depends on {v0[wbs]}\n"
                f"\t- Avilable valid 2d bsplines: {lok_bs}\n"
            )
            raise Exception(msg)

        dbs[k0] = coll.ddata[v0][wbs][0]
        dref[k0] = coll.ddata[v0]['ref']

    # -----------------
    # check uniformity
    # -----------------

    # bsplines
    lbs = list(set(dbs.values()))
    if len(lbs) != 1:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dbs.items()]
        msg = (
            "Line tracing for vector field: non-uniform bsplines 2d!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # ref
    lref = list(set(dref.values()))
    if len(lref) != 1:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dref.items()]
        msg = (
            "Line tracing for vector field: non-uniform ref!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # -------------------------
    # common bsplines and mesh

    dkey.update({
        'key_mesh': coll.dobj[wbs][lbs[0]][wm],
        'key_bs': lbs[0],
        'ref': lref[0],
        'ref_bs': coll.dobj[wbs][lbs[0]]['ref'],
    })

    return dkey


# #############################################################
# #############################################################
#                  Check pts
# #############################################################


def _check_pts(
    pts_X=None,
    pts_Y=None,
    pts_Z=None,
):

    # -------------
    # initialize
    # -------------

    dpts = {'pts_X': pts_X, 'pts_Y': pts_Y, 'pts_Z': pts_Z}

    # -------------
    # type and finite
    # -------------

    for k0, v0 in dpts.items():

        # convert
        if np.isscalar(v0):
            dpts[k0] = np.r_[v0]
        elif isinstance(v0, (list, tuple)):
            dpts[k0] = np.array(v0)

        # check
        if not isinstance(dpts[k0], np.ndarray):
            msg = (
                f"Line tracing for vector field, arg '{k0}' wrong type!\n"
                "\t- expected: np.ndarray\n"
                f"\t- Provided: {type(dpts[k0])}\n"
            )
            raise Exception(msg)

        # should have at least one finite value
        if not np.any(np.isfinite(dpts[k0])):
            msg = (
            )
            raise Exception(msg)

    # -------------
    # uniformity
    # -------------

    # shape
    dshape = {k0: v0.shape for k0, v0 in dpts.items()}

    if len(set(dshape.values())) != 1:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dshape.items()]
        msg = (
            "Line tracing for vector field, "
            "pts coordinates have different shape!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # iok
    iok = np.all(
        [np.isfinite(v0) for v0 in dpts.values()],
        axis=0,
    )

    if not np.any(iok):
        msg = (
            "Not a single common finite value found in pts_X, pts_Y, ptsZ!\n"
        )
        raise Exception(msg)

    dpts['iok'] = iok

    return dpts


# #############################################################
# #############################################################
#                  Check resolution, length
# #############################################################


def _check_res_length(
    res=None,
    length=None,
    npts=None,
    dl=None,
):

    # ------------------
    # res vs length vs npts
    # ------------------

    lc = [
        res is not None and length is not None,
        res is not None and npts is not None,
        npts is not None and length is not None,
        dl is not None,
    ]
    if np.sum(lc) != 1:
        lnames = [
            ('res', 'length'), ('res', 'npts'),
            ('npts', 'length'), ('dl', '')
        ]
        lstr = [
            f"\t- '{k0}' and '{k1}':\t{lc[ii]}"
            for ii, (k0, k1) in enumerate(lnames)
        ]
        msg = (
            "Field line tracing: provide one only of the following input:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # ------------------
    # check each
    # ------------------

    # res
    if res is not None:
        res = float(ds._generic_check._check_var(
            res, 'res',
            types=(int, float),
            sign='>0.',
        ))

    # npts
    if npts is not None:
        npts = int(ds._generic_check._check_var(
            npts, 'npts',
            types=(int, float),
            sign='>0.',
        ))

    # length
    if length is not None:
        length = float(ds._generic_check._check_var(
            length, 'length',
            types=(int, float),
            sign='>0.',
        ))

    # dl
    if dl is not None:
        dl = ds._generic_check._check_flat1darray(
            dl, 'dl',
            dtypes=float,
            unique=True,
        )

    # ------------------
    # case by case
    # ------------------

    if lc[0]:
        dl = np.linspace(0, length, int(np.ceil(length / res)))

    elif lc[1]:
        dl = np.linspace(0, res * (npts - 1), npts)

    elif lc[2]:
        dl = np.linspace(0, length, npts)

    return dl


# #############################################################
# #############################################################
#              check dsolver
# #############################################################


def _check_dsolver(dsolver=None):

    # ---------------
    # Default
    # ---------------

    ddef = copy.deepcopy(_DSOLVER)

    if dsolver is None:
        dsolver = ddef

    # ---------------
    # check
    # ---------------

    if not isinstance(dsolver, dict):
        _err_dsolver(dsolver, ddef)

    for k0, v0 in ddef.items():
        if dsolver.get(k0) is None:
            dsolver[k0] = v0

    if any([k0 not in ddef.keys() for k0 in dsolver.keys()]):
        _err_dsolver(dsolver, ddef)

    return dsolver


def _err_dsolver(dsolver, ddef):
    lstr = [f"\t- {k0}: {v0} (default)" for k0, v0 in ddef.items()]
    msg = (
        "Arg dsolver must be a dict with some of the following:\n"
        + "\n".join(lstr)
        + "\nIs fed to scipy.integrate.solve_ivp()"
    )

    raise Exception(msg)


# #############################################################
# #############################################################
#              prepare output
# #############################################################


def _prepare_output(
    coll=None,
    dkey=None,
    dpts=None,
    dl=None,
    direct=None,
):

    # -------------
    # npts_full
    # -------------

    # npts_full vs direct
    npts = dl.size
    if len(direct) == 1:
        npts_full = npts
    else:
        npts_full = 2*npts - 1

    # ref_npts
    ref_npts = True

    # -------------
    # ref and shape
    # -------------

    # ref
    ref_key = dkey['ref']
    ref_bs = dkey['ref_bs']
    ref_key_nobs = tuple([rr for rr in ref_key if rr not in ref_bs])

    ibs = ref_key.index(ref_bs[0])
    ref_out = (
        ref_key[:ibs]
        + (ref_npts,) + tuple([None for ss in dpts['pts_X'].shape])
        + ref_key[ibs+2:]
    )

    ref_ev = (
        ref_key[:ibs]
        + tuple([None for ss in dpts['pts_X'].shape])
        + ref_key[ibs+2:]
    )

    # shape
    shape_key = coll.ddata[dkey['key_XR']]['shape']
    shape_key_nobs = tuple([
        shape_key[ii] for ii, rr in enumerate(ref_key)
        if rr not in ref_bs
    ])

    shape_out = (
        shape_key[:ibs]
        + (npts_full,) + dpts['pts_X'].shape
        + shape_key[ibs+2:]
    )

    shape_ev = (
        shape_key[:ibs]
        + dpts['pts_X'].shape
        + shape_key[ibs+2:]
    )

    # nan-filled
    nan = np.full(shape_out, np.nan)
    nanev = np.full(shape_ev, np.nan)

    # ------------------
    # slices and domain
    # ------------------

    update_sli_dom = _get_update_sli_dom(
        ref_out=ref_out,
        ref_key_nobs=ref_key_nobs,
        direct=direct,
        npts=npts,
    )

    # -------------
    # prepare dl_full
    # -------------

    if len(direct) == 1:
        if 'bck' in direct:
            dl_full = -dl
        else:
            dl_full = dl

    else:
        dl_full = np.r_[-dl[:0:-1], dl]
    assert dl_full.size == npts_full

    # -------------
    # dout
    # -------------

    dout = {
        'pts_X': {
            'data': np.copy(nan),
            'units': 'm',
            'ref': ref_out,
        },
        'pts_Y': {
            'data': np.copy(nan),
            'units': 'm',
            'ref': ref_out,
        },
        'pts_Z': {
            'data': np.copy(nan),
            'units': 'm',
            'ref': ref_out,
        },
        'dl': {
            'data': dl_full,
            'ref': (True,),
            'units': 'm',
            'dim': 'length',
        },
    }

    # forward solver info
    if 'fwd' in direct:
        dout.update({
            'nfev_fwd': {
                'data': np.copy(nanev),
                'units': None,
                'ref': ref_ev,
            },
            'status_fwd': {
                'data': np.copy(nanev),
                'units': None,
                'ref': ref_ev,
            },
        })

    # backward solver info
    if 'bck' in direct:
        dout.update({
            'nfev_bck': {
                'data': np.copy(nanev),
                'units': None,
                'ref': ref_ev,
            },
            'status_bck': {
                'data': np.copy(nanev),
                'units': None,
                'ref': ref_ev,
            },
        })

    return dout, update_sli_dom, shape_key_nobs


def _get_update_sli_dom(
    ref_out=None,
    ref_key_nobs=None,
    direct=None,
    npts=None,
):

    # -------------------
    # prepare ifwd, ibck
    # -------------------

    if len(direct) == 2:
        ifwd = slice(npts-1, None, 1)
        ibck = slice(npts-1, None, -1)

    else:
        if 'fwd' in direct:
            ifwd = slice(None)
            ibck = None

        else:
            ibck = slice(None)
            ifwd = None

    # --------------
    # initialize
    # --------------

    # -------
    # slices

    sli_fwd = [ifwd if rr is True else 0 for rr in ref_out]
    sli_bck = [ibck if rr is True else 0 for rr in ref_out]

    sli_ev = [0 for rr in ref_out[:-1]]

    # ----------
    # indr, indp

    indr = [ii for ii, rr in enumerate(ref_out) if rr in ref_key_nobs]
    indp = [ii for ii, rr in enumerate(ref_out) if rr is None]

    ref_ev = [rr for rr in ref_out if rr is not True]
    indr_ev = [ii for ii, rr in enumerate(ref_ev) if rr in ref_key_nobs]
    indp_ev = [ii for ii, rr in enumerate(ref_ev) if rr is None]

    # -------
    # domain

    domain = {k0: {'ind': None} for k0 in ref_key_nobs}

    # -------------
    # update func
    # -------------

    def func(
        ipts, iref,
        sli_fwd=np.array(sli_fwd),
        sli_bck=np.array(sli_bck),
        sli_ev=np.array(sli_ev),
        indr=np.array(indr, dtype=int),
        indp=np.array(indp, dtype=int),
        indr_ev=np.array(indr_ev, dtype=int),
        indp_ev=np.array(indp_ev, dtype=int),
        domain=domain,
        ref_key_nobs=ref_key_nobs,
    ):

        # --------
        # ev

        sli_ev[indr_ev] = iref
        sli_ev[indp_ev] = ipts

        # --------
        # pts

        if sli_fwd is not None:
            sli_fwd[indr] = iref
            sli_fwd[indp] = ipts

        if sli_bck is not None:
            sli_bck[indr] = iref
            sli_bck[indp] = ipts

        # --------
        # domain

        for ii, k0 in enumerate(ref_key_nobs):
            domain[k0]['ind'] = iref[ii]

        return (
            tuple(sli_fwd), tuple(sli_bck),
            domain,
            tuple(sli_ev),
        )

    return func


# #############################################################
# #############################################################
#              Get derivative functions
# #############################################################


def _get_func(
    coll=None,
    dkey=None,
    geometry=None,
):

    if geometry == 'toroidal':

        func_fwd = _get_func_toroidal(
            coll=coll,
            dkey=dkey,
            sign=1.,
        )

        func_bck = _get_func_toroidal(
            coll=coll,
            dkey=dkey,
            sign=-1.,
        )

    else:

        func_fwd = _get_func_linear(
            coll=coll,
            dkey=dkey,
            sign=1.,
        )

        func_bck = _get_func_linear(
            coll=coll,
            dkey=dkey,
            sign=-1.,
        )

    return func_fwd, func_bck


def _get_func_toroidal(coll=None, dkey=None, sign=1.):

    def func(ll, pxyz, domain, coll=coll, sign=sign, dkey=dkey):

        R = np.hypot(pxyz[0], pxyz[1])

        dout = coll.interpolate(
            keys=[dkey['key_XR'], dkey['key_YZ'], dkey['key_Zphi']],
            ref_key=dkey['key_bs'],
            x0=R,
            x1=pxyz[2],
            grid=False,
            domain=domain,
            details=False,
            store=False,
        )

        # extract and normalized
        BR = dout[dkey['key_XR']]['data']
        BZ = dout[dkey['key_YZ']]['data']
        Bphi = dout[dkey['key_Zphi']]['data']
        BN_sign_inv = sign / np.sqrt(BR**2 + BZ**2 + Bphi**2)

        cosphi = pxyz[0] / R
        sinphi = pxyz[1] / R

        BXn = (BR * cosphi - Bphi * sinphi) * BN_sign_inv
        BYn = (BR * sinphi + Bphi * cosphi) * BN_sign_inv
        BZn = BZ * BN_sign_inv

        return [BXn[0], BYn[0], BZn[0]]

    return func


def _get_func_linear(coll=None, dkey=None, sign=1.):

    def func(ll, pxyz, domain, coll=coll, sign=sign, dkey=dkey):

        dout = coll.interpolate(
            keys=[dkey['key_XR'], dkey['key_YZ'], dkey['key_Zphi']],
            ref_key=dkey['key_bs'],
            x0=pxyz[0],
            x1=pxyz[1],
            grid=False,
            domain=domain,
            details=False,
            store=False,
        )

        # extract and normalized
        BX = dout[dkey['key_XR']]['data']
        BY = dout[dkey['key_YZ']]['data']
        BZ = dout[dkey['key_Zphi']]['data']
        BN_sign_inv = sign / np.sqrt(BX**2 + BY**2 + BZ**2)

        BXn = BX * BN_sign_inv
        BYn = BY * BN_sign_inv
        BZn = BZ * BN_sign_inv

        return [BXn[0], BYn[0], BZn[0]]

    return func


# #############################################################
# #############################################################
#              add dphi
# #############################################################


def _add_dphi(
    dout=None,
    direct=None,
    npts=None,
):

    # ------------
    # axis
    # ------------

    axis = dout['pts_X']['ref'].index(True)

    # ------------
    # phi
    # ------------

    phi = np.arctan2(dout['pts_Y']['data'], dout['pts_X']['data'])

    # ------------
    # get nturns
    # ------------

    if len(direct) == 2:
        i0 = npts - 1

        sli_bck = tuple([
            slice(i0, None, -1) if ii == axis else slice(None)
            for ii in range(phi.ndim)
        ])
        sli_fwd = tuple([
            slice(i0, None, 1) if ii == axis else slice(None)
            for ii in range(phi.ndim)
        ])
        nturns_fwd = _add_turns(phi[sli_fwd], axis=axis)
        nturns_bck = _add_turns(phi[sli_bck], axis=axis)

        sli_rev = tuple([
            slice(None, 0, -1) if ii == axis else slice(None)
            for ii in range(phi.ndim)
        ])
        nturns = np.concatenate(
            (nturns_bck[sli_rev], nturns_fwd),
            axis=axis,
        )

    else:
        i0 = 0
        nturns = _add_turns(phi, axis=axis)

    # -----------------
    # derive dphi
    # -----------------

    sli0 = tuple([
        slice(i0, i0+1) if ii == axis else slice(None)
        for ii in range(phi.ndim)
    ])
    dphi = (phi - phi[sli0]) + nturns * 2*np.pi

    # ------------
    # store
    # ------------

    dout.update({
        'dphi': {
            'data': dphi,
            'units': 'rad',
            'ref': dout['pts_X']['ref'],
            'dim': 'angle',
        }
    })

    return


def _add_turns(phi, axis=None):

    dphi = np.diff(phi, axis=axis)

    sli0 = tuple([
        slice(0, 1) if ii == axis else slice(None)
        for ii in range(dphi.ndim)
    ])
    sign = np.sign(dphi[sli0])

    ind = np.concatenate(
        (np.zeros(sign.shape), dphi * sign < 0),
        axis=axis,
    )

    return sign * np.cumsum(ind, axis=axis)


# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
from matplotlib.path import Path
import datastock as ds


# specific
from . import _class01_select as _select


# ##################################################################
# ##################################################################
#                          crop rect mesh
# ##################################################################


def crop(
    coll=None,
    key=None,
    crop=None,
    thresh_in=None,
    remove_isolated=None,
):
    """ Crop a rect mesh

    Parameters
    ----------
    key:        str
        key of the rect mesh to be cropped
    crop:      np.ndarray
        Can be either:
            - bool: a boolean mask array
            - float: a closed 2d polygon used for cropping
    threshin:   int
        minimum nb. of corners for a mesh element to be included
    remove_isolated: bool
        flag indicating whether to remove isolated mesh elements

    Return
    ------
    crop:       np.ndarray
        bool mask
    key:        str
        key of the rect mesh to be cropped
    thresh_in:  int
        minimum nb. of corners for a mesh element to be included

    """

    # ------------
    # check inputs
    # ------------

    key, mtype, cropbool, thresh_in, remove_isolated = _crop_check(
        coll=coll,
        key=key,
        crop=crop,
        thresh_in=thresh_in,
        remove_isolated=remove_isolated,
    )

    # -----------
    # if crop is a poly => compute as bool
    # -----------

    if not cropbool:

        # --------------
        # rect

        if mtype == 'rect':

            (Rc, Zc), (Rk, Zk) = coll.select_mesh_elements(
                key=key,
                elements='cents',
                return_neighbours=True,
                returnas='data',
            )

            nR, nZ = Rc.shape
            npts = Rk.shape[-1] + 1

            pts = np.concatenate(
                (
                    np.concatenate((Rc[:, :, None], Rk), axis=-1)[..., None],
                    np.concatenate((Zc[:, :, None], Zk), axis=-1)[..., None],
                ),
                axis=-1,
            ).reshape((npts*nR*nZ, 2))

            isin = Path(crop.T).contains_points(pts).reshape((nR, nZ, npts))
            crop = np.sum(isin, axis=-1) >= thresh_in

            # Remove isolated pixelsi
            if remove_isolated is True:
                # All pixels should have at least one neighbour in R and one in Z
                # This constraint is useful for discrete gradient evaluation (D1N2)
                crop0 = crop
                while True:

                    # neighR
                    neighR = np.copy(crop0)
                    neighR[0, :] &= neighR[1, :]
                    neighR[-1, :] &= neighR[-2, :]
                    neighR[1:-1, :] &= (neighR[:-2, :] | neighR[2:, :])

                    # neighZ
                    neighZ = np.copy(crop0)
                    neighZ[:, 0] &= neighZ[:, 1]
                    neighZ[:, -1] &= neighZ[:, -2]
                    neighZ[:, 1:-1] &= (neighZ[:, :-2] | neighZ[:, 2:])

                    # overall
                    crop = neighR & neighZ

                    # stop or continue
                    if np.all(crop[crop0]):
                        break
                    else:
                        crop0 = crop

        # --------------
        # tri

        else:

            Rc, Zc = coll.select_mesh_elements(
                key=key,
                elements='cents',
                return_neighbours=False,
                returnas='data',
            )

            pts = np.array([Rc, Zc]).T
            crop = Path(crop.T).contains_points(pts)

    return crop, key, thresh_in


# #############################################################################
# #############################################################################
#                           check
# #############################################################################


def _crop_check(
    coll=None,
    key=None,
    crop=None,
    thresh_in=None,
    remove_isolated=None,
):

    # -----------
    # key
    # -----------

    wm = coll._which_mesh
    lkm = list(coll.dobj[wm].keys())
    key = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lkm,
    )
    meshtype = coll.dobj[wm][key]['type']

    if meshtype not in ['rect', 'tri']:
        raise NotImplementedError()

    # shape
    shape = coll.dobj[wm][key]['shape_c']

    # -----------
    # crop
    # -----------

    # basic
    c0 = (
        isinstance(crop, np.ndarray)
        and crop.ndim == 2
        and np.all(np.isfinite(crop))
        and (
            (
                crop.shape[0] == 2
                and np.allclose(crop[:, 0], crop[:, -1])
                and crop.dtype in [int, float]
            )
            or (
                crop.shape == shape
                and crop.dtype == bool
            )
        )
    )
    if not c0:
        msg = (
            "Arg crop must be either:\n"
            f"\t- array of bool: mask of shape {shape}\n"
            f"\t- array of floats: (2, npts) closed (R, Z) polygon\n"
            f"Provided:\n{crop}"
        )
        raise Exception(msg)

    cropbool = crop.dtype == np.bool_

    # --------------------
    # thresh_in and maxth
    # --------------------

    maxth = 5 if coll.dobj[wm][key]['type'] == 'rect' else 4

    thresh_in = int(ds._generic_check._check_var(
        thresh_in, 'thresh_in',
        default=3,
        types=(int, float),
        sign=['>=1', f"<={maxth}"],
    ))

    # ----------------
    # remove_isolated
    # ----------------

    remove_isolated = ds._generic_check._check_var(
        remove_isolated, 'remove_isolated',
        default=True,
        types=bool,
    )

    return key, meshtype, cropbool, thresh_in, remove_isolated


# #############################################################################
# #############################################################################
#                           crop bsplines
# #############################################################################


def add_cropbs_from_crop(coll=None, keybs=None, keym=None):

    # ----------------
    # get

    kcropbs = False
    wm = coll._which_mesh
    wbs = coll._which_bsplines

    if coll.dobj[wm][keym]['crop'] is not False:
        kcropm = coll.dobj[wm][keym]['crop']
        cropbs = _get_cropbs_from_crop(
            coll=coll,
            crop=coll.ddata[kcropm]['data'],
            keybs=keybs,
        )
        kcropbs = f'{keybs}_crop'
        kcroppedbs = f'{keybs}_nbs_crop'

    elif coll.dobj[wm][keym]['type'] == 'tri':
        kcropbs = f'{keybs}_crop'

    # ----------------
    # optional crop

    if kcropbs is not False:

        # add cropped flat reference
        coll.add_ref(
            key=kcroppedbs,
            size=int(cropbs.sum()),
        )

        coll.add_data(
            key=kcropbs,
            data=cropbs,
            ref=coll._dobj[wbs][keybs]['ref'],
            dim='bool',
            quant='bool',
        )
        coll._dobj[wbs][keybs]['crop'] = kcropbs


# #############################################################################
# #############################################################################
#                           Mesh2DRect - crop
# #############################################################################


def _get_cropbs_from_crop(coll=None, crop=None, keybs=None):

    if isinstance(crop, str) and crop in coll.ddata.keys():
        crop = coll.ddata[crop]['data']

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    shref = coll.dobj[wm][coll.dobj[wbs][keybs][wm]]['shape_c']
    if crop.shape != shref:
        msg = "Arg crop seems to have the wrong shape!"
        raise Exception(msg)

    keym = coll.dobj[wbs][keybs][coll._which_mesh]
    kRk, kZk = coll.dobj[wm][keym]['knots']
    kRc, kZc = coll.dobj[wm][keym]['cents']

    cents_per_bs_R, cents_per_bs_Z = _select._mesh2DRect_bsplines_knotscents(
        returnas='ind',
        return_knots=False,
        return_cents=True,
        ind=None,
        deg=coll.dobj[wbs][keybs]['deg'],
        Rknots=coll.ddata[kRk]['data'],
        Zknots=coll.ddata[kZk]['data'],
        Rcents=coll.ddata[kRc]['data'],
        Zcents=coll.ddata[kZc]['data'],
    )

    shapebs = coll.dobj[wbs][keybs]['shape']
    cropbs = np.array([
        [
            np.all(crop[cents_per_bs_R[:, ii], cents_per_bs_Z[:, jj]])
            for jj in range(shapebs[1])
        ]
        for ii in range(shapebs[0])
    ], dtype=bool)

    return cropbs



# -*- coding: utf-8 -*-
"""
Created on Wed Dec  6 08:25:06 2023

@author: dvezinet
"""


import datastock as ds


# ############################################################
# ############################################################
#                 BSplines - saving
# ############################################################


def prepare_bsplines(coll=None):

    # -----------------
    # Remove classes

    wbs = coll._which_bsplines

    dclas = {}
    lkbs = list(coll.dobj.get(wbs, {}).keys())
    for kbs in lkbs:
        dclas[kbs] = coll._dobj[wbs][kbs]['class']
        coll._dobj[wbs][kbs]['class'] = None

    return dclas


def restore_bsplines(coll=None, dclas=None):

    # -----------------
    # Remove classes

    wbs = coll._which_bsplines

    lkbs = list(coll.dobj.get(wbs, {}).keys())
    for kbs in lkbs:
        coll._dobj[wbs][kbs]['class'] = dclas[kbs]

    return


# ############################################################
# ############################################################
#                 BSplines - loading
# ############################################################


def load(
    pfe=None,
    cls=None,
    coll=None,
    allow_pickle=None,
    sep=None,
    verb=None,
):

    # --------------------
    # use datastock.load()

    from . import _class02_compute as _compute

    if cls is None:
        from ._class03_Bins import Bins
        cls = Bins

    coll = ds.load(
        pfe=pfe,
        cls=cls,
        coll=coll,
        allow_pickle=allow_pickle,
        sep=sep,
        verb=verb,
    )

    # ----------------
    # re-build classes

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    lkbs = list(coll.dobj.get(wbs, {}).keys())
    for kbs in lkbs:

        keym = coll.dobj[wbs][kbs][wm]
        nd = coll.dobj[wm][keym]['nd']
        mtype = coll.dobj[wm][keym]['type']
        deg = coll.dobj[wbs][kbs]['deg']
        if nd == '1d':
            dref, ddata, dobj = _compute._mesh1d_bsplines(
                coll=coll, keym=keym, keybs=kbs, deg=deg,
            )
        elif mtype == 'rect':
            dref, ddata, dobj = _compute._mesh2DRect_bsplines(
                coll=coll, keym=keym, keybs=kbs, deg=deg,
            )
        elif mtype == 'tri':
            dref, ddata, dobj = _compute._mesh2DTri_bsplines(
                coll=coll, keym=keym, keybs=kbs, deg=deg,
            )
        else:
            dref, ddata, dobj = _compute._mesh2Dpolar_bsplines(
                coll=coll, keym=keym, keybs=kbs, deg=deg, # angle=angle,
            )

        coll._dobj[wbs][kbs]['class'] = dobj[wbs][kbs]['class']

    return coll


# ###############
# __version__
# ###############


from . import _version
__version__ = _version.version
__version_tuple__ = _version.version_tuple


# ######################
# sub-packages
# ######################


from ._class03_Bins import Bins as BSplines2D
from ._saveload import load
from . import tests



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds
import matplotlib.pyplot as plt
from matplotlib.path import Path


# tofu
from . import _generic_mesh


# ##############################################################
# ##############################################################
#                       Main
# ##############################################################


def main(
    coll=None,
    key=None,
    res_RZ=None,
    mode=None,
    res_phi=None,
):

    # -------------
    # check inputs

    (
        key, res_phi,
    ) = _check(
        coll=coll,
        key=key,
        res_phi=res_phi,
    )

    # ---------------------
    # prepare dsamp
    # ---------------------

    dsamp = coll.get_sample_mesh(
        key=key,
        res=res_RZ,
        grid=False,
        mode=mode,
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        in_mesh=True,
        store=False,
    )

    x0u = dsamp['x0']['data']
    x1u = dsamp['x1']['data']
    index = dsamp['ind']['data']

    # ------------------
    # phi from res_phi
    # ------------------

    nphi = np.ceil(3 + x0u * (2*np.pi) / res_phi).astype(int)

    # -----------------
    # func_pts_from_ind
    # -----------------

    func_RZphi_from_ind = _get_func_RZphi_from_ind(
        x0u=x0u,
        x1u=x1u,
        nphi=nphi,
    )

    # -----------------
    # func_pts_from_domain
    # -----------------

    func_ind_from_domain = _get_func_ind_from_domain(
        x0u=x0u,
        x1u=x1u,
        index=index,
        nphi=nphi,
        # debug
        func_RZphi_from_ind=func_RZphi_from_ind,
    )

    return func_RZphi_from_ind, func_ind_from_domain


# ##########################################################
# ##########################################################
#               checks
# ##########################################################


def _check(
    coll=None,
    key=None,
    res_phi=None,
):

    # ---------
    # mesh
    # ---------

    # key
    wm = coll._which_mesh
    key, _, cat = _generic_mesh._get_key_mesh_vs_bplines(
        coll=coll,
        key=key,
        which=wm,
    )

    lok = ['rect', 'tri']
    if coll.dobj[wm][key]['type'] not in lok:
        msg = (
            f"Only accepts mtype in {lok}\n"
            f"Provided: {key}\n"
        )
        raise Exception(msg)

    # ---------
    # res_phi
    # ---------

    res_phi = float(ds._generic_check._check_var(
        res_phi, 'res_phi',
        types=(int, float),
        sign='>0',
    ))

    return key, res_phi


# ##########################################################
# ##########################################################
#           pts_from_ind
# ##########################################################


def _get_func_RZphi_from_ind(
    x0u=None,
    x1u=None,
    nphi=None,
):

    # --------------
    # func
    # -------------

    def func(
        indr=None,
        indz=None,
        indphi=None,
        # resources
        x0u=x0u,
        x1u=x1u,
        nphi=nphi,
    ):

        # ----------
        # check
        # ----------

        if indphi is not None:
            c0 = (indr is not None) and (indr.shape == indphi.shape)
            if not c0:
                indr_str = 'None' if indr is None else indr.shape
                msg = (
                    "Arg indr and indphi must be the same shape!\n"
                    f"Provided:\n"
                    f"\t- indr.shape = {indr_str}\n"
                    f"\t- indphi.shape = {indphi.shape}\n"
                )
                raise Exception(msg)

        # ----------
        # R, Z
        # ----------

        # R
        if indr is not None:
            R = x0u[indr]
        else:
            R = None

        # Z
        if indz is not None:
            Z = x1u[indz]
        else:
            Z = None

        # dS
        dS = (x0u[1] - x0u[0]) * (x1u[1] - x1u[0])

        # -------------
        # phi
        # -------------

        dV = None
        phi = None
        if indr is not None:
            dV = np.full(indr.shape, np.nan)
            iru = np.unique(indr)

            if indphi is None:
                for iri in iru:
                    phii = _get_phi_from_nphi(nphi[iri])

                    iok = indr == iri
                    dV[iok] = dS * x0u[iri] * (phii[1] - phii[0])

            else:
                phi = np.full(indr.shape, np.nan)
                for iri in iru:
                    phii = _get_phi_from_nphi(nphi[iri])

                    iok = indr == iri
                    dV[iok] = dS * x0u[iri] * (phii[1] - phii[0])
                    phi[iok] = phii[indphi[iok]]

        return R, Z, phi, dV

    return func


# ##########################################################
# ##########################################################
#           pts_ind_from_domain
# ##########################################################


def _get_func_ind_from_domain(
    x0u=None,
    x1u=None,
    nphi=None,
    index=None,
    # debug
    func_RZphi_from_ind=None,
):

    x0f = np.repeat(x0u[:, None], x1u.size, axis=1)
    x1f = np.repeat(x1u[None, :], x0u.size, axis=0)
    ptsf = np.array([x0f.ravel(), x1f.ravel()]).T
    shapef = x0f.shape

    # ---------------------
    # get polygons - cross
    # ---------------------

    # ---------------------
    # get cross-section polygon with margin

    def func(
        # from domain
        DR=None,
        DZ=None,
        Dphi=None,
        # from poly
        pcross0=None,
        pcross1=None,
        phor0=None,
        phor1=None,
        # resources
        index=index,
        x0u=x0u,
        x1u=x1u,
        ptsf=ptsf,
        shapef=shapef,
        nphi=nphi,
        # debug
        debug=None,
        debug_msg=None,
        func_RZphi_from_ind=func_RZphi_from_ind,
    ):
        # ------------------
        # check
        # ------------------

        index = np.copy(index)

        DR, DZ, Dphi = _check_domain(
            DR=DR,
            DZ=DZ,
            Dphi=Dphi,
        )

        # ------------------
        # limits from domain
        # ------------------

        if DR is not None:
            index &= (x0u[:, None] >= DR[0]) & (x0u[:, None] <= DR[1])
        if DZ is not None:
            index &= (x1u[None, :] >= DZ[0]) & (x1u[None, :] <= DZ[1])

        # ------------------
        # limits from pcross
        # ------------------

        if pcross0 is not None:
            pcross = Path(np.array([pcross0, pcross1]).T)
            index &= pcross.contains_points(ptsf).reshape(shapef)

        # R and Z indices
        ir, iz = index.nonzero()
        iru = np.unique(ir)

        # ------------
        # indices
        # ------------

        ind = np.empty((3, 0), dtype=int)
        for ii, iri in enumerate(iru):

            # ---
            # iz

            izi = np.unique(iz[ir == iri])
            if izi.size == 0:
                print(f"SKIP: {ii} {iri}")
                continue

            # -----
            # iphi

            if phor0 is None and Dphi is None:
                iphi = np.arange(0, nphi[iri])

            elif Dphi is not None:
                phii = _get_phi_from_nphi(nphi[iri])
                if Dphi[0] < Dphi[1]:
                    iphi = np.nonzero(
                        (phii >= Dphi[0]) & (phii <= Dphi[1])
                    )[0]
                else:
                    iphi = np.nonzero(
                        (phii >= Dphi[0]) | (phii <= Dphi[1])
                    )[0]
            elif phor0 is not None:
                phii = np.pi*np.linspace(-1, 1, nphi[iri])
                pts = np.array([
                    x0u[iri]*np.cos(phii),
                    x0u[iri]*np.sin(phii),
                ]).T
                path = Path(np.array([phor0, phor1]).T)
                iphi = path.contains_points(pts).nonzero()[0]

            # ---------
            # group ind

            indi = np.array(
                [
                    np.full((iphi.size*izi.size,), iri),
                    np.repeat(izi, iphi.size),
                    np.tile(iphi, izi.size),
                ],
                dtype=int,
            )

            ind = np.concatenate((ind, indi), axis=1)

        # ----------
        # debug

        if debug is True:

            # coordinates
            rr, zz, pp, dV = func_RZphi_from_ind(ind)

            # title
            tit = (
                "func_ind_from_domain\n"
                f"pcross0 {pcross0 is not None} - phor0 {phor0 is not None}\n"
                f"DR {DR is not None} - DZ {DZ is not None}"
                f" - Dphi{Dphi is not None}"
            )
            if debug_msg is not None:
                tit += f"\n{debug_msg}"

            # figure and axes
            fig = plt.figure(figsize=(14, 8))
            fig.suptitle(tit, size=14, fontweight='bold')

            ax0 = fig.add_subplot(1, 2, 1, aspect='equal')
            ax0.set_xlabel("R (m)", size=12, fontweight='bold')
            ax0.set_ylabel("Z (m)", size=12, fontweight='bold')

            ax1 = fig.add_subplot(1, 2, 2, aspect='equal')
            ax1.set_xlabel("X (m)", size=12, fontweight='bold')
            ax1.set_ylabel("Y (m)", size=12, fontweight='bold')

            # domain
            if DR is not None:
                ax0.axvline(DR[0], c='k', ls='--')
                ax0.axvline(DR[1], c='k', ls='--')

            if DZ is not None:
                ax0.axhline(DZ[0], c='k', ls='--')
                ax0.axhline(DZ[1], c='k', ls='--')

            if Dphi is not None:
                maxr = np.max(rr)
                ax1.plot(
                    np.r_[0, maxr]*np.cos(Dphi[0]),
                    np.r_[0, maxr]*np.sin(Dphi[0]),
                    c='k',
                    ls='--',
                )
                ax1.plot(
                    np.r_[0, maxr]*np.cos(Dphi[1]),
                    np.r_[0, maxr]*np.sin(Dphi[1]),
                    c='k',
                    ls='--',
                )

            # polygons
            if pcross0 is not None:
                ax0.fill(pcross0, pcross1, fc=(0.5, 0.5, 0.5, 0.5))
            if phor0 is not None:
                ax1.fill(phor0, phor1, fc=(0.5, 0.5, 0.5, 0.5))

            # points
            ax0.plot(rr, zz, '.')
            ax1.plot(rr*np.cos(pp), rr*np.sin(pp), '.')

        return ind[0, :], ind[1, :], ind[2, :]

    return func


# ##########################################################
# ##########################################################
#           unique get phi function
# ##########################################################


def _get_phi_from_nphi(nphi):
    return np.pi * np.linspace(-1, 1, nphi, endpoint=False)


# ##########################################################
# ##########################################################
#           _check_domain
# ##########################################################


def _check_domain(
    DR=None,
    DZ=None,
    Dphi=None,
):
    # -----------
    #   DR
    # -----------

    if DR is not None:
        DR = ds._generic_check._check_flat1darray(
            DR, 'DR',
            size=2,
            dtype=float,
            unique=True,
            sign='>=0',
        )

    # -----------
    #   DZ
    # -----------

    if DZ is not None:
        DZ = ds._generic_check._check_flat1darray(
            DZ, 'DZ',
            size=2,
            dtype=float,
            unique=True,
        )

    # -----------
    #   Dphi
    # -----------

    if Dphi is not None:
        Dphi = ds._generic_check._check_flat1darray(
            Dphi, 'Dphi',
            size=2,
            dtype=float,
        )
        Dphi = [
            np.arctan2(np.sin(Dphi[0]), np.cos(Dphi[0])),
            np.arctan2(np.sin(Dphi[1]), np.cos(Dphi[1])),
        ]

    return DR, DZ, Dphi



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec


# ##################################################
# ##################################################
#                       Main
# ##################################################


def main(
    coll=None,
    key=None,
    res=None,
    # slice
    Z=None,
    phi=None,
    # domain
    DR=None,
    DZ=None,
    Dphi=None,
    # option
    reshape_2d=None,
    adjust_phi=None,
    # plot
    plot=None,
    dax=None,
    color=None,
):

    # --------------
    # check inputs
    # --------------

    (
        res, Z, phi, domain, reshape_2d, adjust_phi, plot,
    ) = _check(
        res=res,
        # slice
        Z=Z,
        phi=phi,
        # domain
        DR=DR,
        DZ=DZ,
        Dphi=Dphi,
        # option
        reshape_2d=reshape_2d,
        adjust_phi=adjust_phi,
        plot=plot,
    )

    # --------------
    # prepare
    # --------------

    (
        func_RZphi_from_ind,
        func_ind_from_domain,
    ) = coll.get_sample_mesh_3d_func(
        key=key,
        res_RZ=res,
        res_phi=res,
        mode='abs',
    )

    # --------------
    # compute
    # --------------

    if phi is None:
        indr, indz, indphi = _horizontal_slice(
            res=res,
            # sampling
            func_ind_from_domain=func_ind_from_domain,
            func_RZphi_from_ind=func_RZphi_from_ind,
            # slice
            Z=Z,
            # domain
            **domain,
        )

    else:
        indr, indz, indphi = _poloidal_slice(
            res=res,
            # sampling
            func_ind_from_domain=func_ind_from_domain,
            func_RZphi_from_ind=func_RZphi_from_ind,
            # slice
            phi=phi,
            # option
            reshape_2d=reshape_2d,
            # domain
            **domain,
        )

    # ------------
    # get points
    # ------------

    pts_r, pts_z, pts_phi, _ = func_RZphi_from_ind(
        indr=indr,
        indz=indz,
        indphi=indphi,
    )

    # ------------
    # adjust phi
    # ------------

    if adjust_phi is True:
        pts_phi = phi

    # --------------
    # output
    # --------------

    dout = {
        # indices
        'indr': {
            'data': indr,
            'units': 'index',
        },
        'indz': {
            'data': indz,
            'units': 'index',
        },
        'indphi': {
            'data': indphi,
            'units': 'index',
        },
        # points
        'pts_r': {
            'data': pts_r,
            'units': 'm',
        },
        'pts_z': {
            'data': pts_z,
            'units': 'm',
        },
        'pts_phi': {
            'data': pts_phi,
            'units': 'rad',
        },
    }

    # -----------------
    # optional plotting
    # -----------------

    if plot is True:
        _plot(
            dout,
            dax=dax,
            color=color,
        )

    return dout


# ##############################################
# ##############################################
#                   Check
# ##############################################


def _check(
    res=None,
    # slice
    Z=None,
    phi=None,
    # domain
    DR=None,
    DZ=None,
    Dphi=None,
    # option
    reshape_2d=None,
    adjust_phi=None,
    # plot
    plot=None,
    color=None,
):

    # ---------
    # res
    # ---------

    res = float(ds._generic_check._check_var(
        res, 'res',
        types=(int, float),
        sign=">0",
    ))

    # ---------
    # Z vs phi
    # ---------

    lc = [
        Z is not None,
        phi is not None,
    ]
    if np.sum(lc) != 1:
        msg = (
            "Please provide Z xor phi!\n"
            "\t- Z: height of horizontal slice\n"
            "\t- phi: toroidal angle of poloidal slice\n"
            "Provided:\n"
            f"\t- Z = {Z}\n"
            f"\t- phi = {phi}\n"
        )
        raise Exception(msg)

    if phi is None:
        Z = float(ds._generic_check._check_var(
            Z, 'Z',
            types=(int, float),
        ))
    else:
        phi = float(ds._generic_check._check_var(
            phi, 'phi',
            types=(int, float),
        ))
        phi = np.arctan2(np.sin(phi), np.cos(phi))

    # ---------
    # domain
    # ---------

    domain = {
        'DR': DR,
        'DZ': DZ,
        'Dphi': Dphi,
    }

    for k0, v0 in domain.items():
        if v0 is not None:
            v0 = ds._generic_check._check_flat1darray(
                v0, k0,
                dtype=float,
                unique=True,
                size=2,
            )

            if v0[0] >= v0[1]:
                msg = (
                    f"If provided, arg '{k0}' must be strictly increasing!\n"
                    + "For Dphi, use +2pi if needed\n" if k0 == 'Dphi' else ''
                )
                raise Exception(msg)

    if phi is None:
        del domain['DZ']
    else:
        del domain['Dphi']

    # ---------
    # reshape_2d
    # ---------

    reshape_2d = ds._generic_check._check_var(
        reshape_2d, 'reshape_2d',
        types=bool,
        default=True,
    )

    # ---------
    # adjust_phi
    # ---------

    adjust_phi = ds._generic_check._check_var(
        adjust_phi, 'adjust_phi',
        types=bool,
        default=False,
    )
    if phi is None:
        adjust_phi = False

    # ---------
    # plot
    # ---------

    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=False,
    )

    return res, Z, phi, domain, reshape_2d, adjust_phi, plot


# ##############################################
# ##############################################
#               Horizontal slice
# ##############################################


def _horizontal_slice(
    res=None,
    # pre-sampling
    func_ind_from_domain=None,
    func_RZphi_from_ind=None,
    # slice
    Z=None,
    # domain
    DR=None,
    Dphi=None,
):
    # -----------
    # domain Z
    # -----------

    dZ = 1.5*res
    DZ = (Z - dZ, Z + dZ)

    indr, indz, indphi = func_ind_from_domain(
        DR=DR,
        DZ=DZ,
        Dphi=Dphi,
    )

    # ------------
    # select plane
    # ------------

    izu = np.unique(indz)
    _, zz, _, _ = func_RZphi_from_ind(indz=izu)

    iz = np.argmin(np.abs(zz - Z))
    iok = indz == izu[iz]

    return indr[iok], indz[iok], indphi[iok]


# ##############################################
# ##############################################
#               Poloidal slice
# ##############################################


def _poloidal_slice(
    res=None,
    # pre-sampling
    func_ind_from_domain=None,
    func_RZphi_from_ind=None,
    # slice
    phi=None,
    # domain
    DR=None,
    DZ=None,
    # option
    reshape_2d=None,
):
    # -----------
    # domain Z
    # -----------

    dphi = np.pi/12
    Dphi = (phi - dphi, phi + dphi)

    indr, indz, indphi = func_ind_from_domain(
        DR=DR,
        DZ=DZ,
        Dphi=Dphi,
    )

    # ------------
    # select plane
    # ------------

    iru = np.unique(indr)
    iphi = np.zeros(iru.shape)
    indr_new, indz_new, indphi_new = [], [], []
    for ii, iri in enumerate(iru):
        ind = indr == iri
        rr, _, phii, dV = func_RZphi_from_ind(
            indr=indr[ind],
            indphi=indphi[ind]
        )

        iphi[ii] = indphi[ind][np.argmin(np.abs(phii - phi))]

        ind[ind] = indphi[ind] == iphi[ii]
        indr_new.append(indr[ind])
        indz_new.append(indz[ind])
        indphi_new.append(indphi[ind])

    # ---------
    # reshape
    # ---------

    ir = np.concatenate(indr_new)
    iz = np.concatenate(indz_new)
    iphi = np.concatenate(indphi_new)

    # ------------
    # safety check

    assert np.unique([ir, iz], axis=1).shape[1] == ir.size

    # ----------------
    # optional reshape

    if reshape_2d is True:
        i0u = np.unique(ir)
        i1u = np.unique(iz)

        indr = -np.ones((i0u.size, i1u.size), dtype=int)
        indz = -np.ones((i0u.size, i1u.size), dtype=int)
        indphi = -np.ones((i0u.size, i1u.size), dtype=int)

        for ii, iri in enumerate(i0u):
            ind = ir == iri

            indsz = np.argsort(iz[ind])
            iiz = np.searchsorted(i1u, np.sort(iz[ind]))
            sli = (ii, iiz)

            indr[sli] = iri
            indz[sli] = iz[ind][indsz]
            indphi[sli] = iphi[ind][indsz]

    else:
        indr = ir
        indz = iz
        indphi = iphi

    return indr, indz, indphi


# ##############################################
# ##############################################
#               Plot
# ##############################################


def _plot(
    dout=None,
    dax=None,
    color=None,
):
    # --------------
    # prepare data
    # --------------

    iok = dout['indr']['data'] >= 0
    xx = dout['pts_r']['data'][iok] * np.cos(dout['pts_phi']['data'][iok])
    yy = dout['pts_r']['data'][iok] * np.sin(dout['pts_phi']['data'][iok])

    # --------------
    # prepare figure
    # --------------

    if dax is None:

        dmargin = {
            'left': 0.05, 'right': 0.95,
            'bottom': 0.06, 'top': 0.90,
            'hspace': 0.20, 'wspace': 0.20,
        }

        fig = plt.figure(figsize=(14, 8))
        gs = gridspec.GridSpec(ncols=3, nrows=1, **dmargin)

        # --------------
        # prepare axes

        ax0 = fig.add_subplot(gs[:, 0], aspect='equal', adjustable='box')
        ax0.set_ylabel('Z (m)', size=12, fontweight='bold')
        ax0.set_xlabel('R (m)', size=12, fontweight='bold')
        ax0.set_title("cross-section", size=14, fontweight='bold')

        ax1 = fig.add_subplot(gs[:, 1], aspect='equal', adjustable='box')
        ax1.set_xlabel('X (m)', size=12, fontweight='bold')
        ax1.set_ylabel('Y (m)', size=12, fontweight='bold')
        ax1.set_title("horizontal", size=14, fontweight='bold')

        ax2 = fig.add_subplot(
            gs[:, 2],
            aspect='equal',
            adjustable='box',
            projection='3d',
        )
        ax2.set_xlabel('X (m)', size=12, fontweight='bold')
        ax2.set_ylabel('Y (m)', size=12, fontweight='bold')
        ax2.set_zlabel('Z (m)', size=12, fontweight='bold')
        ax2.set_title("3d", size=14, fontweight='bold')

        dax = {
            'cross': {'handle': ax0},
            'hor': {'handle': ax1},
            '3d': {'handle': ax2},
        }

    # --------------
    # plot cross
    # --------------

    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.plot(
            dout['pts_r']['data'][iok],
            dout['pts_z']['data'][iok],
            marker='.',
            linestyle='None',
            ms=6,
            color=color,
        )

    # --------------
    # plot hor
    # --------------

    kax = 'hor'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.plot(
            xx,
            yy,
            marker='.',
            linestyle='None',
            ms=6,
            color=color,
        )

    # --------------
    # plot 3d
    # --------------

    kax = '3d'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.plot(
            xx,
            yy,
            dout['pts_z']['data'][iok],
            marker='.',
            linestyle='None',
            ms=6,
            color=color,
        )

    return
